def shorten(s, n=100):
    """Shorten string s to at most n characters, appending "..." if necessary."""
    if s is None:
        return None
    if len(s) > n:
        s = s[:n-3] + '...'
    return s



import io
from xopen import xopen


class FileWriter:
    def __init__(self, file, _close_file=None):
        self._file = file
        if isinstance(file, str):
            self._file = xopen(file, 'wb')
            self._close_on_exit = True
        else:
            self._close_on_exit = bool(_close_file)

    def close(self):
        if self._close_on_exit:
            self._file.close()

    def __enter__(self):
        if self._file.closed:
            raise ValueError("I/O operation on closed file")
        return self

    def __exit__(self, *args):
        self.close()


class FastaWriter(FileWriter):
    """
    Write FASTA-formatted sequences to a file.
    """

    def __init__(self, file, line_length=None, _close_file=None):
        """
        If line_length is not None, the lines will
        be wrapped after line_length characters.
        """
        super().__init__(file, _close_file=_close_file)
        self.line_length = line_length if line_length != 0 else None

    def write(self, name_or_record, sequence=None):
        """Write an entry to the the FASTA file.

        If only one parameter (name_or_record) is given, it must have
        attributes .name and .sequence, which are then used.
        Otherwise, the first parameter must be the name and the second
        the sequence.

        The effect is that you can write this:
        writer.write("name", "ACCAT")
        or
        writer.write(Sequence("name", "ACCAT"))
        """
        if sequence is None:
            name = name_or_record.name
            sequence = name_or_record.sequence
        else:
            name = name_or_record

        if self.line_length is not None:
            self._file.write(('>' + name + '\n').encode('ascii'))
            s = []
            for i in range(0, len(sequence), self.line_length):
                s.append(sequence[i:i + self.line_length] + '\n')
            self._file.write(''.join(s).encode('ascii'))
        else:
            s = '>' + name + '\n' + sequence + '\n'
            self._file.write(s.encode('ascii'))


class FastqWriter(FileWriter):
    """
    Write sequences with qualities in FASTQ format.

    FASTQ files are formatted like this:
    @read name
    SEQUENCE
    +
    QUALITIS
    """
    file_mode = 'wb'

    def __init__(self, file, two_headers=False, _close_file=None):
        super().__init__(file, _close_file=_close_file)
        self._two_headers = two_headers

    def write(self, record):
        """
        Write a Sequence record to the FASTQ file.

        The record object must have attributes .name, .sequence and .qualities.
        """
        name2 = record.name if self._two_headers else ''
        s = ('@' + record.name + '\n' + record.sequence + '\n+'
             + name2 + '\n' + record.qualities + '\n')
        self._file.write(s.encode('ascii'))

    def writeseq(self, name, sequence, qualities):
        self._file.write("@{0:s}\n{1:s}\n+\n{2:s}\n".format(
            name, sequence, qualities).encode('ascii'))



# cython: language_level=3, emit_code_comments=False

from libc.string cimport strncmp
cimport cython

from .exceptions import FastqFormatError
from ._util import shorten


cdef class Sequence:
	"""
	A record in a FASTA or FASTQ file. For FASTA, the qualities attribute
	is None. For FASTQ, qualities is a string and it contains the qualities
	encoded as ascii(qual+33).
	"""
	cdef:
		public str name
		public str sequence
		public str qualities

	def __cinit__(self, str name, str sequence, str qualities=None):
		"""Set qualities to None if there are no quality values"""
		self.name = name
		self.sequence = sequence
		self.qualities = qualities

		if qualities is not None and len(qualities) != len(sequence):
			rname = shorten(name)
			raise ValueError("In read named {!r}: length of quality sequence "
				"({}) and length of read ({}) do not match".format(
					rname, len(qualities), len(sequence)))

	def __getitem__(self, key):
		"""slicing"""
		return self.__class__(
			self.name,
			self.sequence[key],
			self.qualities[key] if self.qualities is not None else None)

	def __repr__(self):
		qstr = ''
		if self.qualities is not None:
			qstr = ', qualities={!r}'.format(shorten(self.qualities))
		return '<Sequence(name={!r}, sequence={!r}{})>'.format(
			shorten(self.name), shorten(self.sequence), qstr)

	def __len__(self):
		return len(self.sequence)

	def __richcmp__(self, other, int op):
		if 2 <= op <= 3:
			eq = self.name == other.name and \
				self.sequence == other.sequence and \
				self.qualities == other.qualities
			if op == 2:
				return eq
			else:
				return not eq
		else:
			raise NotImplementedError()

	def __reduce__(self):
		return (Sequence, (self.name, self.sequence, self.qualities))


# It would be nice to be able to have the first parameter be an
# unsigned char[:] (memory view), but this fails with a BufferError
# when a bytes object is passed in.
# See <https://stackoverflow.com/questions/28203670/>

ctypedef fused bytes_or_bytearray:
	bytes
	bytearray


def paired_fastq_heads(bytes_or_bytearray buf1, bytes_or_bytearray buf2, Py_ssize_t end1, Py_ssize_t end2):
	"""
	Skip forward in the two buffers by multiples of four lines.

	Return a tuple (length1, length2) such that buf1[:length1] and
	buf2[:length2] contain the same number of lines (where the
	line number is divisible by four).
	"""
	cdef:
		Py_ssize_t pos1 = 0, pos2 = 0
		Py_ssize_t linebreaks = 0
		unsigned char* data1 = buf1
		unsigned char* data2 = buf2
		Py_ssize_t record_start1 = 0
		Py_ssize_t record_start2 = 0

	while True:
		while pos1 < end1 and data1[pos1] != b'\n':
			pos1 += 1
		if pos1 == end1:
			break
		pos1 += 1
		while pos2 < end2 and data2[pos2] != b'\n':
			pos2 += 1
		if pos2 == end2:
			break
		pos2 += 1
		linebreaks += 1
		if linebreaks == 4:
			linebreaks = 0
			record_start1 = pos1
			record_start2 = pos2

	# Hit the end of the data block
	return record_start1, record_start2


def fastq_iter(file, sequence_class, Py_ssize_t buffer_size):
	"""
	Parse a FASTQ file and yield Sequence objects

	The *first value* that the generator yields is a boolean indicating whether
	the first record in the FASTQ has a repeated header (in the third row
	after the ``+``).

	file -- a file-like object, opened in binary mode (it must have a readinto
	method)

	buffer_size -- size of the initial buffer. This is automatically grown
	    if a FASTQ record is encountered that does not fit.
	"""
	cdef:
		bytearray buf = bytearray(buffer_size)
		char[:] buf_view = buf
		char* c_buf = buf
		int endskip
		str name
		char* name_encoded
		Py_ssize_t bufstart, bufend, pos, record_start, sequence_start
		Py_ssize_t second_header_start, sequence_length, qualities_start
		Py_ssize_t second_header_length, name_length
		bint custom_class = sequence_class is not Sequence
		Py_ssize_t n_records = 0
		bint extra_newline = False

	if buffer_size < 1:
		raise ValueError("Starting buffer size too small")

	# buf is a byte buffer that is re-used in each iteration. Its layout is:
	#
	# |-- complete records --|
	# +---+------------------+---------+-------+
	# |   |                  |         |       |
	# +---+------------------+---------+-------+
	# ^   ^                  ^         ^       ^
	# 0   bufstart           end       bufend  len(buf)
	#
	# buf[0:bufstart] is the 'leftover' data that could not be processed
	# in the previous iteration because it contained an incomplete
	# FASTQ record.

	readinto = file.readinto
	bufstart = 0

	# The input file is processed in chunks that each fit into buf
	while True:
		assert bufstart < len(buf_view)
		bufend = readinto(buf_view[bufstart:]) + bufstart
		if bufstart == bufend:
			# End of file
			if bufstart > 0 and buf_view[bufstart-1] != b'\n':
				# There is still data in the buffer and its last character is
				# not a newline: This is a file that is missing the final
				# newline. Append a newline and continue.
				buf_view[bufstart] = b'\n'
				bufstart += 1
				bufend += 1
				extra_newline = True
			else:
				break

		# Parse all complete FASTQ records in this chunk
		pos = 0
		record_start = 0
		while True:
			# Parse the name (line 0)
			if c_buf[pos] != b'@':
				raise FastqFormatError("Line expected to "
					"start with '@', but found {!r}".format(chr(c_buf[pos])),
					line=n_records * 4)
			pos += 1
			while pos < bufend and c_buf[pos] != b'\n':
				pos += 1
			if pos == bufend:
				break
			endskip = 1 if c_buf[pos-1] == b'\r' else 0
			name_length = pos - endskip - record_start - 1
			name_encoded = c_buf + record_start + 1
			# .decode('latin-1') is 50% faster than .decode('ascii')
			name = c_buf[record_start+1:pos-endskip].decode('latin-1')

			pos += 1

			# Parse the sequence (line 1)
			sequence_start = pos
			while pos < bufend and c_buf[pos] != b'\n':
				pos += 1
			if pos == bufend:
				break
			endskip = 1 if c_buf[pos-1] == b'\r' else 0
			sequence = c_buf[sequence_start:pos-endskip].decode('latin-1')
			sequence_length = pos - endskip - sequence_start
			pos += 1

			# Parse second header (line 2)
			second_header_start = pos
			if pos == bufend:
				break
			if c_buf[pos] != b'+':
				raise FastqFormatError("Line expected to "
					"start with '+', but found {!r}".format(chr(c_buf[pos])),
					line=n_records * 4 + 2)
			pos += 1  # skip over the '+'
			while pos < bufend and c_buf[pos] != b'\n':
				pos += 1
			if pos == bufend:
				break
			endskip = 1 if c_buf[pos-1] == b'\r' else 0
			second_header_length = pos - endskip - second_header_start - 1
			if second_header_length == 0:
				second_header = False
			else:
				if (name_length != second_header_length or
						strncmp(c_buf+second_header_start+1,
							name_encoded, second_header_length) != 0):
					raise FastqFormatError(
						"Sequence descriptions don't match ('{}' != '{}').\n"
						"The second sequence description must be either "
						"empty or equal to the first description.".format(
							name_encoded[:name_length].decode('latin-1'),
							c_buf[second_header_start+1:pos-endskip]
							.decode('latin-1')), line=n_records * 4 + 2)
				second_header = True
			pos += 1

			# Parse qualities (line 3)
			qualities_start = pos
			while pos < bufend and c_buf[pos] != b'\n':
				pos += 1
			if pos == bufend:
				break
			endskip = 1 if c_buf[pos-1] == b'\r' else 0
			qualities = c_buf[qualities_start:pos-endskip].decode('latin-1')
			if pos - endskip - qualities_start != sequence_length:
				raise FastqFormatError("Length of sequence and "
					"qualities differ", line=n_records * 4 + 3)
			pos += 1
			if n_records == 0:
				yield second_header  # first yielded value is special
			if custom_class:
				yield sequence_class(name, sequence, qualities)
			else:
				yield Sequence.__new__(Sequence, name, sequence, qualities)
			n_records += 1
			record_start = pos
			if pos == bufend:
				break
		if pos == bufend:
			if record_start == 0 and bufend == len(buf):
				# buffer too small, double it
				buffer_size *= 2
				prev_buf = buf
				buf = bytearray(buffer_size)
				buf[0:bufend] = prev_buf
				del prev_buf
				bufstart = bufend
				buf_view = buf
				c_buf = buf
			else:
				bufstart = bufend - record_start
				buf[0:bufstart] = buf[record_start:bufend]
	if pos > record_start:
		if extra_newline:
			pos -= 1
		lines = buf[record_start:pos].count(b'\n')
		raise FastqFormatError(
			'Premature end of file encountered. The incomplete final record was: '
			'{!r}'.format(shorten(buf[record_start:pos].decode('latin-1'), 500)),
			line=n_records * 4 + lines)



"""
Sequence I/O: Read and write FASTA and FASTQ files efficiently
"""

__all__ = [
    'open',
    'Sequence',
    'FastaReader',
    'FastaWriter',
    'FastqReader',
    'FastqWriter',
    'UnknownFileFormat',
    'FileFormatError',
    'FastaFormatError',
    'FastqFormatError',
    'InterleavedSequenceReader',
    'InterleavedSequenceWriter',
    'PairedSequenceReader',
]

import os
from contextlib import ExitStack
import functools

from xopen import xopen

from ._core import Sequence
from .readers import FastaReader, FastqReader
from .writers import FastaWriter, FastqWriter
from .exceptions import UnknownFileFormat, FileFormatError, FastaFormatError, FastqFormatError
from .chunks import read_chunks, read_paired_chunks

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


def open(file1, *, file2=None, fileformat=None, interleaved=False, mode='r', qualities=None):
    """
    Open sequence files in FASTA or FASTQ format for reading or writing. This is
    a factory that returns an instance of one of the ...Reader or ...Writer
    classes also defined in this module.

    file1, file2 -- Paths to regular or compressed files or file-like
        objects. Use file1 if data is single-end. If also file2 is provided,
        sequences are paired.

    mode -- Either 'r' for reading or 'w' for writing.

    interleaved -- If True, then file1 contains interleaved paired-end data.
        file2 must be None in this case.

    fileformat -- If set to None, the file format is autodetected from the file name
        extension. Set to 'fasta' or 'fastq' to not auto-detect.

    qualities -- When mode is 'w' and fileformat is None, this can be set to
        True or False to specify whether the written sequences will have quality
        values. This is is used in two ways:
        * If the output format cannot be determined (unrecognized extension
          etc), no exception is raised, but fasta or fastq format is chosen
          appropriately.
        * When False (no qualities available), an exception is raised when the
          auto-detected output format is FASTQ.
    """
    if mode not in ('r', 'w'):
        raise ValueError("Mode must be 'r' or 'w'")
    if interleaved and file2 is not None:
        raise ValueError("When interleaved is set, file2 must be None")
    if file2 is not None:
        if mode == 'r':
            return PairedSequenceReader(file1, file2, fileformat)
        else:
            return PairedSequenceWriter(file1, file2, fileformat, qualities)
    if interleaved:
        if mode == 'r':
            return InterleavedSequenceReader(file1, fileformat)
        else:
            return InterleavedSequenceWriter(file1, fileformat, qualities)

    # The multi-file options have been dealt with, delegate rest to the
    # single-file function.
    return _open_single(
        file1, fileformat=fileformat, mode=mode, qualities=qualities)


def _detect_format_from_name(name):
    """
    name -- file name

    Return 'fasta', 'fastq' or None if the format could not be detected.
    """
    name = name.lower()
    for ext in ('.gz', '.xz', '.bz2'):
        if name.endswith(ext):
            name = name[:-len(ext)]
            break
    name, ext = os.path.splitext(name)
    if ext in ['.fasta', '.fa', '.fna', '.csfasta', '.csfa']:
        return 'fasta'
    elif ext in ['.fastq', '.fq'] or (ext == '.txt' and name.endswith('_sequence')):
        return 'fastq'
    return None


def _open_single(file, *, fileformat=None, mode='r', qualities=None):
    """
    Open a single sequence file. See description of open() above.
    """
    if mode not in ('r', 'w'):
        raise ValueError("Mode must be 'r' or 'w'")
    if isinstance(file, str):
        file = xopen(file, mode + 'b')
        close_file = True
    else:
        if mode == 'r' and not hasattr(file, 'readinto'):
            raise ValueError(
                'When passing in an open file-like object, it must have been opened in binary mode')
        close_file = False
    if mode == 'r':
        fastq_handler = FastqReader
        fasta_handler = FastaReader
    else:
        fastq_handler = FastqWriter
        fasta_handler = FastaWriter
    fastq_handler = functools.partial(fastq_handler, _close_file=close_file)
    fasta_handler = functools.partial(fasta_handler, _close_file=close_file)

    if fileformat:  # Explict file format given
        fileformat = fileformat.lower()
        if fileformat == 'fasta':
            return fasta_handler(file)
        elif fileformat == 'fastq':
            return fastq_handler(file)
        else:
            raise UnknownFileFormat(
                "File format {!r} is unknown (expected 'fasta' or 'fastq').".format(fileformat))

    # First, try to detect the file format from the file name only
    format = None
    if hasattr(file, "name") and isinstance(file.name, str):
        format = _detect_format_from_name(file.name)
    if format is None and mode == 'w' and qualities is not None:
        # Format not recognized, but we know whether to use a format with or without qualities
        format = 'fastq' if qualities else 'fasta'

    if mode == 'r' and format is None:
        # No format detected so far. Try to read from the file.
        if file.seekable():
            first_char = file.read(1)
            file.seek(-1, 1)
        else:
            first_char = file.peek(1)[0:1]
        if first_char == b'#':
            # A comment char - only valid for some FASTA variants (csfasta)
            format = 'fasta'
        elif first_char == b'>':
            format = 'fasta'
        elif first_char == b'@':
            format = 'fastq'
        elif first_char == b'':
            # Empty input. Pretend this is FASTQ
            format = 'fastq'
        else:
            raise UnknownFileFormat(
                'Could not determine whether file {!r} is FASTA or FASTQ. The file extension was '
                'not available or not recognized and the first character in the file ({!r}) is '
                'unexpected.'.format(file, first_char))

    if format is None:
        assert mode == 'w'
        raise UnknownFileFormat('Cannot determine whether to write in FASTA or FASTQ format')

    if format == 'fastq' and mode == 'w' and qualities is False:
        raise ValueError(
            'Output format cannot be FASTQ since no quality values are available.')

    return fastq_handler(file) if format == 'fastq' else fasta_handler(file)


def _sequence_names_match(r1, r2):
    """
    Check whether the sequence records r1 and r2 have identical names, ignoring a
    suffix of '1' or '2'. Some old paired-end reads have names that end in '/1'
    and '/2'. Also, the fastq-dump tool (used for converting SRA files to FASTQ)
    appends a .1 and .2 to paired-end reads if option -I is used.
    """
    name1 = r1.name.split(None, 1)[0]
    name2 = r2.name.split(None, 1)[0]
    if name1[-1:] in '12' and name2[-1:] in '12':
        name1 = name1[:-1]
        name2 = name2[:-1]
    return name1 == name2


class PairedSequenceReader:
    """
    Read paired-end reads from two files.

    Wraps two BinaryFileReader instances, making sure that reads are properly
    paired.
    """
    paired = True

    def __init__(self, file1, file2, fileformat=None):
        with ExitStack() as stack:
            self.reader1 = stack.enter_context(_open_single(file1, fileformat=fileformat))
            self.reader2 = stack.enter_context(_open_single(file2, fileformat=fileformat))
            self._close = stack.pop_all().close
        self.delivers_qualities = self.reader1.delivers_qualities

    def __iter__(self):
        """
        Iterate over the paired reads. Each item is a pair of Sequence objects.
        """
        # Avoid usage of zip() below since it will consume one item too many.
        it1, it2 = iter(self.reader1), iter(self.reader2)
        while True:
            try:
                r1 = next(it1)
            except StopIteration:
                # End of file 1. Make sure that file 2 is also at end.
                try:
                    next(it2)
                    raise FileFormatError("Reads are improperly paired. There are more reads in "
                        "file 2 than in file 1.", line=None) from None
                except StopIteration:
                    pass
                break
            try:
                r2 = next(it2)
            except StopIteration:
                raise FileFormatError("Reads are improperly paired. There are more reads in "
                    "file 1 than in file 2.", line=None) from None
            if not _sequence_names_match(r1, r2):
                raise FileFormatError("Reads are improperly paired. Read name '{}' "
                    "in file 1 does not match '{}' in file 2.".format(r1.name, r2.name), line=None) from None
            yield (r1, r2)

    def close(self):
        self._close()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()


class InterleavedSequenceReader:
    """
    Read paired-end reads from an interleaved FASTQ file.
    """
    paired = True

    def __init__(self, file, fileformat=None):
        self.reader = _open_single(file, fileformat=fileformat)
        self.delivers_qualities = self.reader.delivers_qualities

    def __iter__(self):
        it = iter(self.reader)
        for r1 in it:
            try:
                r2 = next(it)
            except StopIteration:
                raise FileFormatError("Interleaved input file incomplete: Last record "
                    "{!r} has no partner.".format(r1.name), line=None) from None
            if not _sequence_names_match(r1, r2):
                raise FileFormatError("Reads are improperly paired. Name {!r} "
                    "(first) does not match {!r} (second).".format(r1.name, r2.name), line=None)
            yield (r1, r2)

    def close(self):
        self.reader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class PairedSequenceWriter:
    def __init__(self, file1, file2, fileformat='fastq', qualities=None):
        with ExitStack() as stack:
            self._writer1 = stack.enter_context(_open_single(file1, fileformat=fileformat, mode='w',
                qualities=qualities))
            self._writer2 = stack.enter_context(_open_single(file2, fileformat=fileformat, mode='w',
                qualities=qualities))
            self._close = stack.pop_all().close

    def write(self, read1, read2):
        self._writer1.write(read1)
        self._writer2.write(read2)

    def close(self):
        self._close()

    def __enter__(self):
        # TODO do not allow this twice
        return self

    def __exit__(self, *args):
        self.close()


class InterleavedSequenceWriter:
    """
    Write paired-end reads to an interleaved FASTA or FASTQ file
    """

    def __init__(self, file, fileformat='fastq', qualities=None):

        self._writer = _open_single(
            file, fileformat=fileformat, mode='w', qualities=qualities)

    def write(self, read1, read2):
        self._writer.write(read1)
        self._writer.write(read2)

    def close(self):
        self._writer.close()

    def __enter__(self):
        # TODO do not allow this twice
        return self

    def __exit__(self, *args):
        self.close()
